require 'socket'
require 'selenium-webdriver'
class  Applitools::Driver

  include Selenium::WebDriver::DriverExtensions::HasInputDevices

  attr_reader :remote_server_url, :remote_session_id, :user_agent, :screenshot_taker
  attr_accessor :user_inputs, :browser

  DEFAULT_DRIVER = :firefox
  DRIVER_METHODS = [
    :title, :execute_script, :execute_async_script, :quit, :close, :get,
    :post, :page_source, :window_handles, :window_handle, :switch_to, 
    :navigate, :manage, :capabilities
  ]

  ## If driver is not provided, Applitools::Driver will default to Firefox driver
  ## Driver param can be a Selenium::WebDriver or a named symbol (:chrome)
  #
  ## Example:
  #  eyes.open(browser: :chrome) ##=> will create chrome webdriver
  #  eyes.open(browser: Selenium::WebDriver.for(:chrome) ##=> will create the same thing
  #  eyes.open ##=> will create a webdriver according to Applitools::Driver::DEFAULT_DRIVER
  def initialize(options={})
    browser_obj = options.delete(:browser) || DEFAULT_DRIVER 
    @browser ||= case browser_obj
                 when Symbol
                   Selenium::WebDriver.for browser_obj
                 else
                   browser_obj
                 end
    at_exit { quit rescue nil }

    @user_inputs = []
    @remote_server_url = address_of_remote_server
    @remote_session_id = remote_session_id
    @user_agent = get_user_agent
    begin
      if browser.capabilities.takes_screenshot?
       @screenshot_taker = false
      else
        @screenshot_taker = Applitools::ScreenshotTaker.new(@remote_server_url, @remote_session_id)
      end
    rescue => e 
      raise Applitools::EyesError.new "Can't take screenshots (#{e.message})"
    end
  end

  DRIVER_METHODS.each do |method|
    define_method method do |*args, &block|
      browser.send(method,*args, &block)
    end
  end

  def screenshot_as(output_type)
    return browser.screenshot_as(output_type) if !screenshot_taker

    if output_type.downcase.to_sym != :base64
      raise Applitools::EyesError.new("#{output_type} ouput type not supported for screenshot")
    end
    screenshot_taker.screenshot
  end 

  def mouse
    Applitools::EyesMouse.new(self, browser.mouse)
  end

  def keyboard
    Applitools::EyesKeyboard.new(self, browser.keyboard)
  end

  def find_element(by, selector)
    Applitools::Element.new(self, browser.find_element(by, selector))
  end

  def find_elements(by, selector)
    browser.find_elements(by, selector).map { |el| Applitools::Element.new(self, el) }
  end

  def create_application
    Applitools::TargetApp.new(remote_server_url, remote_session_id, user_agent)
  end

  def clear_user_inputs
    user_inputs.clear
  end

  def ie?
    browser.to_s == 'ie'
  end

  def firefox?
    browser.to_s == 'firefox'
  end

  private

    def address_of_remote_server
      uri = URI(browser.current_url)
      raise Applitools::EyesError.new("Failed to get remote web driver url") if uri.to_s.empty?

      webdriver_host = uri.host
      if ['127.0.0.1', 'localhost'].include?(webdriver_host) && !firefox? && !ie?
        uri.host = get_local_ip || 'localhost'
      end

      uri
    end

    def remote_session_id
      browser.remote_session_id
    end

    def get_user_agent
      execute_script "return navigator.userAgent"
    rescue => e
      puts "getUserAgent(): Failed to obtain user-agent string (#{e.message})"
    end
    
    def get_local_ip
      begin
        Socket.ip_address_list.detect do |intf| 
          intf.ipv4? and !intf.ipv4_loopback? and !intf.ipv4_multicast?
        end.ip_address
      rescue SocketError => e
        raise Applitools::EyesError.new("Failed to get local IP! (#{e})")
      end
    end

end

## .bridge, .session_id and .server_url are private methods in Selenium::WebDriver gem
module Selenium::WebDriver
  class Driver
    def remote_session_id
      bridge.session_id
    end
  end

  class Remote::Http::Common
    def get_server_url
      server_url
    end
  end
end
