class Applitools::Eyes
  DEFAULT_MATCH_TIMEOUT = 2.0
  AGENT_ID = 'eyes.selenium.ruby'

  class << self
    attr_accessor :config
  end

  @config = {
    server_url: 'https://eyes.applitools.com',
    user: AGENT_ID
    # apikey: 'please supply an apikey in your initializer'
  }

  attr_reader :agent_connector, :disabled
  attr_accessor :app_name, :test_name, :session, :is_open, :aborted, :viewport_size, :match_timeout, 
                :failure_reports, :test_batch, :match_level, :driver, :host_os, :host_app, :branch_name, :parent_branch_name, 
                :should_match_window_run_once_on_timeout, :session_start_info, :match_window_task

  def config
    self.class.config
  end

  def initialize(params={})
    raise "Please supply an apikey: Eyes.config[:apikey] = ..." unless config[:apikey]

    @disabled = params[:disabled]

    @driver = create_driver(params)
    return if disabled?

    @agent_connector = Applitools::AgentConnector.new(config[:server_url], config[:user], config[:apikey])
    @match_timeout = DEFAULT_MATCH_TIMEOUT
    @failure_reports = Applitools::FailureReports::ON_CLOSE
  end

  def create_driver(params)
    Applitools::Driver.new(browser: params.fetch(:browser, nil))
  end

  def open(params={})
    return driver if disabled?

    if open?
      abort_if_not_closed
      msg = 'a test is alread running'
      EyesLogger.info(msg) and raise Applitools::EyesError.new(msg)
    end

    self.app_name = params.fetch(:app_name)
    self.failure_reports = params.fetch(:failure_reports, self.failure_reports)
    self.match_level = params.fetch(:match_level,MatchLevel::EXACT)
    self.test_name = params.fetch(:test_name)
    self.viewport_size = params.fetch(:viewport_size, nil)

    self.is_open = true
    driver
  end

  def open?
    self.is_open
  end

  def check_window(tag)
    return if disabled?
    raise Applitools::EyesError.new("Eyes not open") if !open?
    if !session
      start_session
      self.match_window_task = Applitools::MatchWindowTask.new(agent_connector, session, driver, match_timeout)
    end

    as_expected = match_window_task.match_window(tag, should_match_window_run_once_on_timeout)
    if !as_expected
      self.should_match_window_run_once_on_timeout = true
      if !session.new_session?
        #EyesLogger.info %( "mismatch! #{ tag ? "" : "(#{tag})" } )
        if failure_reports.to_i == Applitools::FailureReports::IMMEDIATE
          raise Applitools::TestFailedError.new("Mismatch found in '#{start_info.scenario_id_or_name}'"\
                                                " of '#{start_info.app_id_or_name}'")
        end
      end
    end
  end

  def close
    return if disabled?
    self.is_open = false
    return Applitools::TestResults.new if !session

    session_results_url = session.url
    results = agent_connector.stop_session(session,false)
    new_session = session.new_session?
    self.session = nil

    if new_session
      instructions = "Please approve the new baseline at #{session_results_url}"
      EyesLogger.info "--- New test ended.  #{instructions}"
      message = "' #{session_start_info.scenario_id_or_name} of"\
                " #{session_start_info.app_id_or_name}. #{instructions}"
      raise Applitools::NewTestError.new(message, results)
    elsif 0 < results.mismatches || 0 < results.missing 
      EyesLogger.info "--- Failed test ended. See details at #{session_results_url}"
      message = "' #{session_start_info.scenario_id_or_name} of"\
                " #{session_start_info.app_id_or_name}'. See details at #{session_results_url}"
      raise Applitools::TestFailedError.new(message, results)
    end

    EyesLogger.info "--- Test passed. See details at #{session_results_url}"
  end

  ## Use this method to perform seamless testing with selenium through eyes driver.
  ## Using Selenium methods inside the 'test' block will send the messages to Selenium
  ## after creating the Eyes triggers for them.
  ##
  ## Example:
  #    eyes.test(app_name: 'my app1', test_name: 'my test') do |d|
  #      get "http://www.google.com"
  #      check_window("initial")
  #    end
  def test(params={}, &block)
    begin
      open(params)
      yield(self, driver)
      close
    rescue Applitools::EyesError
    ensure 
      abort_if_not_closed
    end
  end


    def abort_if_not_closed
      return if disabled?
      self.is_open = false
      if session
        begin
          agent_connector.stop_session(session,true)
        rescue Applitools::EyesError => e
          EyesLogger.info "Failed to abort server session -> #{e.message} "
        ensure
          self.session = nil
        end
      end
    end

  private

    def disabled? 
      disabled
    end

    def start_session
      assign_viewport_size
      test_batch ||= Applitools::BatchInfo.new
      app_env = Applitools::Environment.new(host_os, host_app, viewport_size)
      application = driver.create_application
      self.session_start_info = Applitools::StartInfo.new(AGENT_ID,
        app_name, test_name, test_batch, app_env, application, 
        match_level,nil, branch_name, parent_branch_name
      )
      self.session = agent_connector.start_session(session_start_info)
      self.should_match_window_run_once_on_timeout = session.new_session?
    end

    def viewport_size?
      self.viewport_size
    end

    def assign_viewport_size
      if viewport_size?
        self.viewport_size = Applitools::ViewportSize.new(driver, viewport_size)
	self.viewport_size.set
      else
        self.viewport_size =  Applitools::ViewportSize.new(driver).extract_viewport_from_browser!
      end
    end

    def to_hash
      Hash[dimension.each_pair.to_a]
    end
end
